@@ -1,4 +1,4 @@
-#!/usr/bin/env python
+#!/usr/bin/env python3
 #-*- coding: utf-8 -*-
 """
     aztec_code_generator
@@ -14,6 +14,8 @@ import math
 import numbers
 import sys
 import array
+from collections import namedtuple
+from enum import Enum
 
 try:
     from PIL import Image, ImageDraw
@@ -26,44 +28,45 @@ try:
 except ImportError:
     from io import StringIO
 
+Config = namedtuple('Config', ('layers', 'codewords', 'cw_bits', 'bits', 'digits', 'text', 'bytes'))
 
 table = {
-    (15, True): {'layers': 1, 'codewords': 17, 'cw_bits': 6, 'bits': 102, 'digits': 13, 'text': 12, 'bytes': 6},
-    (19, False): {'layers': 1, 'codewords': 21, 'cw_bits': 6, 'bits': 126, 'digits': 18, 'text': 15, 'bytes': 8},
-    (19, True): {'layers': 2, 'codewords': 40, 'cw_bits': 6, 'bits': 240, 'digits': 40, 'text': 33, 'bytes': 19},
-    (23, False): {'layers': 2, 'codewords': 48, 'cw_bits': 6, 'bits': 288, 'digits': 49, 'text': 40, 'bytes': 24},
-    (23, True): {'layers': 3, 'codewords': 51, 'cw_bits': 8, 'bits': 408, 'digits': 70, 'text': 57, 'bytes': 33},
-    (27, False): {'layers': 3, 'codewords': 60, 'cw_bits': 8, 'bits': 480, 'digits': 84, 'text': 68, 'bytes': 40},
-    (27, True): {'layers': 4, 'codewords': 76, 'cw_bits': 8, 'bits': 608, 'digits': 110, 'text': 89, 'bytes': 53},
-    (31, False): {'layers': 4, 'codewords': 88, 'cw_bits': 8, 'bits': 704, 'digits': 128, 'text': 104, 'bytes': 62},
-    (37, False): {'layers': 5, 'codewords': 120, 'cw_bits': 8, 'bits': 960, 'digits': 178, 'text': 144, 'bytes': 87},
-    (41, False): {'layers': 6, 'codewords': 156, 'cw_bits': 8, 'bits': 1248, 'digits': 232, 'text': 187, 'bytes': 114},
-    (45, False): {'layers': 7, 'codewords': 196, 'cw_bits': 8, 'bits': 1568, 'digits': 294, 'text': 236, 'bytes': 145},
-    (49, False): {'layers': 8, 'codewords': 240, 'cw_bits': 8, 'bits': 1920, 'digits': 362, 'text': 291, 'bytes': 179},
-    (53, False): {'layers': 9, 'codewords': 230, 'cw_bits': 10, 'bits': 2300, 'digits': 433, 'text': 348, 'bytes': 214},
-    (57, False): {'layers': 10, 'codewords': 272, 'cw_bits': 10, 'bits': 2720, 'digits': 516, 'text': 414, 'bytes': 256},
-    (61, False): {'layers': 11, 'codewords': 316, 'cw_bits': 10, 'bits': 3160, 'digits': 601, 'text': 482, 'bytes': 298},
-    (67, False): {'layers': 12, 'codewords': 364, 'cw_bits': 10, 'bits': 3640, 'digits': 691, 'text': 554, 'bytes': 343},
-    (71, False): {'layers': 13, 'codewords': 416, 'cw_bits': 10, 'bits': 4160, 'digits': 793, 'text': 636, 'bytes': 394},
-    (75, False): {'layers': 14, 'codewords': 470, 'cw_bits': 10, 'bits': 4700, 'digits': 896, 'text': 718, 'bytes': 446},
-    (79, False): {'layers': 15, 'codewords': 528, 'cw_bits': 10, 'bits': 5280, 'digits': 1008, 'text': 808, 'bytes': 502},
-    (83, False): {'layers': 16, 'codewords': 588, 'cw_bits': 10, 'bits': 5880, 'digits': 1123, 'text': 900, 'bytes': 559},
-    (87, False): {'layers': 17, 'codewords': 652, 'cw_bits': 10, 'bits': 6520, 'digits': 1246, 'text': 998, 'bytes': 621},
-    (91, False): {'layers': 18, 'codewords': 720, 'cw_bits': 10, 'bits': 7200, 'digits': 1378, 'text': 1104, 'bytes': 687},
-    (95, False): {'layers': 19, 'codewords': 790, 'cw_bits': 10, 'bits': 7900, 'digits': 1511, 'text': 1210, 'bytes': 753},
-    (101, False): {'layers': 20, 'codewords': 864, 'cw_bits': 10, 'bits': 8640, 'digits': 1653, 'text': 1324, 'bytes': 824},
-    (105, False): {'layers': 21, 'codewords': 940, 'cw_bits': 10, 'bits': 9400, 'digits': 1801, 'text': 1442, 'bytes': 898},
-    (109, False): {'layers': 22, 'codewords': 1020, 'cw_bits': 10, 'bits': 10200, 'digits': 1956, 'text': 1566, 'bytes': 976},
-    (113, False): {'layers': 23, 'codewords': 920, 'cw_bits': 12, 'bits': 11040, 'digits': 2116, 'text': 1694, 'bytes': 1056},
-    (117, False): {'layers': 24, 'codewords': 992, 'cw_bits': 12, 'bits': 11904, 'digits': 2281, 'text': 1826, 'bytes': 1138},
-    (121, False): {'layers': 25, 'codewords': 1066, 'cw_bits': 12, 'bits': 12792, 'digits': 2452, 'text': 1963, 'bytes': 1224},
-    (125, False): {'layers': 26, 'codewords': 1144, 'cw_bits': 12, 'bits': 13728, 'digits': 2632, 'text': 2107, 'bytes': 1314},
-    (131, False): {'layers': 27, 'codewords': 1224, 'cw_bits': 12, 'bits': 14688, 'digits': 2818, 'text': 2256, 'bytes': 1407},
-    (135, False): {'layers': 28, 'codewords': 1306, 'cw_bits': 12, 'bits': 15672, 'digits': 3007, 'text': 2407, 'bytes': 1501},
-    (139, False): {'layers': 29, 'codewords': 1392, 'cw_bits': 12, 'bits': 16704, 'digits': 3205, 'text': 2565, 'bytes': 1600},
-    (143, False): {'layers': 30, 'codewords': 1480, 'cw_bits': 12, 'bits': 17760, 'digits': 3409, 'text': 2728, 'bytes': 1702},
-    (147, False): {'layers': 31, 'codewords': 1570, 'cw_bits': 12, 'bits': 18840, 'digits': 3616, 'text': 2894, 'bytes': 1806},
-    (151, False): {'layers': 32, 'codewords': 1664, 'cw_bits': 12, 'bits': 19968, 'digits': 3832, 'text': 3067, 'bytes': 1914},
+    (15, True): Config(layers=1, codewords=17, cw_bits=6, bits=102, digits=13, text=12, bytes=6),
+    (19, False): Config(layers=1, codewords=21, cw_bits=6, bits=126, digits=18, text=15, bytes=8),
+    (19, True): Config(layers=2, codewords=40, cw_bits=6, bits=240, digits=40, text=33, bytes=19),
+    (23, False): Config(layers=2, codewords=48, cw_bits=6, bits=288, digits=49, text=40, bytes=24),
+    (23, True): Config(layers=3, codewords=51, cw_bits=8, bits=408, digits=70, text=57, bytes=33),
+    (27, False): Config(layers=3, codewords=60, cw_bits=8, bits=480, digits=84, text=68, bytes=40),
+    (27, True): Config(layers=4, codewords=76, cw_bits=8, bits=608, digits=110, text=89, bytes=53),
+    (31, False): Config(layers=4, codewords=88, cw_bits=8, bits=704, digits=128, text=104, bytes=62),
+    (37, False): Config(layers=5, codewords=120, cw_bits=8, bits=960, digits=178, text=144, bytes=87),
+    (41, False): Config(layers=6, codewords=156, cw_bits=8, bits=1248, digits=232, text=187, bytes=114),
+    (45, False): Config(layers=7, codewords=196, cw_bits=8, bits=1568, digits=294, text=236, bytes=145),
+    (49, False): Config(layers=8, codewords=240, cw_bits=8, bits=1920, digits=362, text=291, bytes=179),
+    (53, False): Config(layers=9, codewords=230, cw_bits=10, bits=2300, digits=433, text=348, bytes=214),
+    (57, False): Config(layers=10, codewords=272, cw_bits=10, bits=2720, digits=516, text=414, bytes=256),
+    (61, False): Config(layers=11, codewords=316, cw_bits=10, bits=3160, digits=601, text=482, bytes=298),
+    (67, False): Config(layers=12, codewords=364, cw_bits=10, bits=3640, digits=691, text=554, bytes=343),
+    (71, False): Config(layers=13, codewords=416, cw_bits=10, bits=4160, digits=793, text=636, bytes=394),
+    (75, False): Config(layers=14, codewords=470, cw_bits=10, bits=4700, digits=896, text=718, bytes=446),
+    (79, False): Config(layers=15, codewords=528, cw_bits=10, bits=5280, digits=1008, text=808, bytes=502),
+    (83, False): Config(layers=16, codewords=588, cw_bits=10, bits=5880, digits=1123, text=900, bytes=559),
+    (87, False): Config(layers=17, codewords=652, cw_bits=10, bits=6520, digits=1246, text=998, bytes=621),
+    (91, False): Config(layers=18, codewords=720, cw_bits=10, bits=7200, digits=1378, text=1104, bytes=687),
+    (95, False): Config(layers=19, codewords=790, cw_bits=10, bits=7900, digits=1511, text=1210, bytes=753),
+    (101, False): Config(layers=20, codewords=864, cw_bits=10, bits=8640, digits=1653, text=1324, bytes=824),
+    (105, False): Config(layers=21, codewords=940, cw_bits=10, bits=9400, digits=1801, text=1442, bytes=898),
+    (109, False): Config(layers=22, codewords=1020, cw_bits=10, bits=10200, digits=1956, text=1566, bytes=976),
+    (113, False): Config(layers=23, codewords=920, cw_bits=12, bits=11040, digits=2116, text=1694, bytes=1056),
+    (117, False): Config(layers=24, codewords=992, cw_bits=12, bits=11904, digits=2281, text=1826, bytes=1138),
+    (121, False): Config(layers=25, codewords=1066, cw_bits=12, bits=12792, digits=2452, text=1963, bytes=1224),
+    (125, False): Config(layers=26, codewords=1144, cw_bits=12, bits=13728, digits=2632, text=2107, bytes=1314),
+    (131, False): Config(layers=27, codewords=1224, cw_bits=12, bits=14688, digits=2818, text=2256, bytes=1407),
+    (135, False): Config(layers=28, codewords=1306, cw_bits=12, bits=15672, digits=3007, text=2407, bytes=1501),
+    (139, False): Config(layers=29, codewords=1392, cw_bits=12, bits=16704, digits=3205, text=2565, bytes=1600),
+    (143, False): Config(layers=30, codewords=1480, cw_bits=12, bits=17760, digits=3409, text=2728, bytes=1702),
+    (147, False): Config(layers=31, codewords=1570, cw_bits=12, bits=18840, digits=3616, text=2894, bytes=1806),
+    (151, False): Config(layers=32, codewords=1664, cw_bits=12, bits=19968, digits=3832, text=3067, bytes=1914),
 }
 
 polynomials = {
@@ -74,90 +77,70 @@ polynomials = {
     12: 4201,
 }
 
+Mode = Enum('Mode', ('UPPER', 'LOWER', 'MIXED', 'PUNCT', 'DIGIT', 'BINARY'))
+Latch = Enum('Latch', Mode.__members__)
+Shift = Enum('Shift', Mode.__members__)
+Misc = Enum('Misc', ('FLG', 'SIZE', 'RESUME'))
+
 code_chars = {
-    'upper': [
-        'P/S', ' ', 'A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'J', 'K', 'L', 'M', 'N', 'O', 'P', 'Q', 'R', 'S',
-        'T', 'U', 'V', 'W', 'X', 'Y', 'Z', 'L/L', 'M/L', 'D/L', 'B/S'
-    ],
-    'lower': [
-        'P/S', ' ', 'a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i', 'j', 'k', 'l', 'm', 'n', 'o', 'p', 'q', 'r', 's',
-        't', 'u', 'v', 'w', 'x', 'y', 'z', 'U/S', 'M/L', 'D/L', 'B/S'
-    ],
-    'mixed': [
-        'P/S', ' ', '\x01', '\x02', '\x03', '\x04', '\x05', '\x06', '\x07', '\x08', '\t', '\n', '\x0b', '\x0c', '\r',
-        '\x1b', '\x1c', '\x1d', '\x1e', '\x1f', '@', '\\', '^', '_', '`', '|', '~', '\x7f', 'L/L', 'U/L', 'P/L', 'B/S'
-    ],
-    'punct': [
-        'FLG(n)', '\r', '\r\n', '. ', ', ', ': ', '!', '"', '#', '$', '%', '&', "'", '(', ')', '*', '+', ',', '-', '.',
-        '/', ':', ';', '<', '=', '>', '?', '[', ']', '{', '}', 'U/L'
-    ],
-    'digit': [
-        'P/S', ' ', '0', '1', '2', '3', '4', '5', '6', '7', '8', '9', ',', '.', 'U/L', 'U/S'
-    ],
+    Mode.UPPER: [Shift.PUNCT] + list(b' ABCDEFGHIJKLMNOPQRSTUVWXYZ') + [Latch.LOWER, Latch.MIXED, Latch.DIGIT, Shift.BINARY],
+    Mode.LOWER: [Shift.PUNCT] + list(b' abcdefghijklmnopqrstuvwxyz') + [Shift.UPPER, Latch.MIXED, Latch.DIGIT, Shift.BINARY],
+    Mode.MIXED: [Shift.PUNCT] + list(b' \x01\x02\x03\x04\x05\x06\x07\x08\t\n\x0b\x0c\r\x1b\x1c\x1d\x1e\x1f@\\^_`|~\x7f') + [Latch.LOWER, Latch.UPPER, Latch.PUNCT, Shift.BINARY],
+    Mode.PUNCT: [Misc.FLG] + list(b'\r') + [b'\r\n', b'. ', b', ', b': '] + list(b'!"#$%&\'()*+,-./:;<=>?[]{}') + [Latch.UPPER],
+    Mode.DIGIT: [Shift.PUNCT] + list(b' 0123456789,.') + [Latch.UPPER, Shift.UPPER],
 }
 
-upper_chars = code_chars['upper'][1:-4]
-lower_chars = code_chars['lower'][1:-4]
-mixed_chars = code_chars['mixed'][1:-4]
-punct_chars = code_chars['punct'][1:-1]
-digit_chars = code_chars['digit'][1:-2]
-punct_2_chars = [pc for pc in punct_chars if len(pc) == 2]
+punct_2_chars = [pc for pc in code_chars[Mode.PUNCT] if isinstance(pc, bytes) and len(pc) == 2]
 
 E = 99999  # some big number
 
 latch_len = {
-    'upper': {
-        'upper': 0, 'lower': 5, 'mixed': 5, 'punct': 10, 'digit': 5, 'binary': 10
+    Mode.UPPER: {
+        Mode.UPPER: 0, Mode.LOWER: 5, Mode.MIXED: 5, Mode.PUNCT: 10, Mode.DIGIT: 5, Mode.BINARY: 10
     },
-    'lower': {
-        'upper': 10, 'lower': 0, 'mixed': 5, 'punct': 10, 'digit': 5, 'binary': 10
+    Mode.LOWER: {
+        Mode.UPPER: 10, Mode.LOWER: 0, Mode.MIXED: 5, Mode.PUNCT: 10, Mode.DIGIT: 5, Mode.BINARY: 10
     },
-    'mixed': {
-        'upper': 5, 'lower': 5, 'mixed': 0, 'punct': 5, 'digit': 10, 'binary': 10
+    Mode.MIXED: {
+        Mode.UPPER: 5, Mode.LOWER: 5, Mode.MIXED: 0, Mode.PUNCT: 5, Mode.DIGIT: 10, Mode.BINARY: 10
     },
-    'punct': {
-        'upper': 5, 'lower': 10, 'mixed': 10, 'punct': 0, 'digit': 10, 'binary': 15
+    Mode.PUNCT: {
+        Mode.UPPER: 5, Mode.LOWER: 10, Mode.MIXED: 10, Mode.PUNCT: 0, Mode.DIGIT: 10, Mode.BINARY: 15
     },
-    'digit': {
-        'upper': 4, 'lower': 9, 'mixed': 9, 'punct': 14, 'digit': 0, 'binary': 14
+    Mode.DIGIT: {
+        Mode.UPPER: 4, Mode.LOWER: 9, Mode.MIXED: 9, Mode.PUNCT: 14, Mode.DIGIT: 0, Mode.BINARY: 14
     },
-    'binary': {
-        'upper': 0, 'lower': 0, 'mixed': 0, 'punct': 0, 'digit': 0, 'binary': 0
+    Mode.BINARY: {
+        Mode.UPPER: 0, Mode.LOWER: 0, Mode.MIXED: 0, Mode.PUNCT: 0, Mode.DIGIT: 0, Mode.BINARY: 0
     },
 }
 
 shift_len = {
-    'upper': {
-        'upper': E, 'lower': E, 'mixed': E, 'punct': 5, 'digit': E, 'binary': E
+    Mode.UPPER: {
+        Mode.UPPER: E, Mode.LOWER: E, Mode.MIXED: E, Mode.PUNCT: 5, Mode.DIGIT: E, Mode.BINARY: E
     },
-    'lower': {
-        'upper': 5, 'lower': E, 'mixed': E, 'punct': 5, 'digit': E, 'binary': E
+    Mode.LOWER: {
+        Mode.UPPER: 5, Mode.LOWER: E, Mode.MIXED: E, Mode.PUNCT: 5, Mode.DIGIT: E, Mode.BINARY: E
     },
-    'mixed': {
-        'upper': E, 'lower': E, 'mixed': E, 'punct': 5, 'digit': E, 'binary': E
+    Mode.MIXED: {
+        Mode.UPPER: E, Mode.LOWER: E, Mode.MIXED: E, Mode.PUNCT: 5, Mode.DIGIT: E, Mode.BINARY: E
     },
-    'punct': {
-        'upper': E, 'lower': E, 'mixed': E, 'punct': E, 'digit': E, 'binary': E
+    Mode.PUNCT: {
+        Mode.UPPER: E, Mode.LOWER: E, Mode.MIXED: E, Mode.PUNCT: E, Mode.DIGIT: E, Mode.BINARY: E
     },
-    'digit': {
-        'upper': 4, 'lower': E, 'mixed': E, 'punct': 4, 'digit': E, 'binary': E
+    Mode.DIGIT: {
+        Mode.UPPER: 4, Mode.LOWER: E, Mode.MIXED: E, Mode.PUNCT: 4, Mode.DIGIT: E, Mode.BINARY: E
     },
-    'binary': {
-        'upper': E, 'lower': E, 'mixed': E, 'punct': E, 'digit': E, 'binary': E
+    Mode.BINARY: {
+        Mode.UPPER: E, Mode.LOWER: E, Mode.MIXED: E, Mode.PUNCT: E, Mode.DIGIT: E, Mode.BINARY: E
     },
 }
 
 char_size = {
-    'upper': 5, 'lower': 5, 'mixed': 5, 'punct': 5, 'digit': 4, 'binary': 8,
+    Mode.UPPER: 5, Mode.LOWER: 5, Mode.MIXED: 5, Mode.PUNCT: 5, Mode.DIGIT: 4, Mode.BINARY: 8,
 }
 
-modes = [
-    'upper', 'lower', 'mixed', 'punct', 'digit', 'binary',
-]
-
-abbr_modes = {
-    'U': 'upper', 'L': 'lower', 'M': 'mixed', 'P': 'punct', 'D': 'digit', 'B': 'binary',
-}
+abbr_modes = {m.name[0]:m for m in Mode}
 
 
 def prod(x, y, log, alog, gf):
@@ -216,151 +199,117 @@ def find_optimal_sequence(data):
 
     TODO: add support of FLG(n) processing
 
-    :param data: data to encode
+    :param data: string or bytes to encode
     :return: optimal sequence
     """
-    back_to = {
-        'upper': 'upper', 'lower': 'upper', 'mixed': 'upper',
-        'punct': 'upper', 'digit': 'upper', 'binary': 'upper'
-    }
-    cur_len = {
-        'upper': 0, 'lower': E, 'mixed': E, 'punct': E, 'digit': E, 'binary': E
-    }
-    cur_seq = {
-        'upper': [], 'lower': [], 'mixed': [], 'punct': [], 'digit': [], 'binary': []
-    }
-    prev_c = ''
+    back_to = {m: Mode.UPPER for m in Mode}
+    cur_len = {m: 0 if m==Mode.UPPER else E for m in Mode}
+    cur_seq = {m: [] for m in Mode}
+    prev_c = None
+    if isinstance(data, str):
+        data = data.encode('iso-8859-1')
     for c in data:
-        for x in modes:
-            for y in modes:
+        for x in Mode:
+            for y in Mode:
                 if cur_len[x] + latch_len[x][y] < cur_len[y]:
                     cur_len[y] = cur_len[x] + latch_len[x][y]
-                    if y == 'binary':
+                    cur_seq[y] = cur_seq[x][:]
+                    back_to[y] = y
+                    if y == Mode.BINARY:
                         # for binary mode use B/S instead of B/L
-                        if x in ['punct', 'digit']:
+                        if x in (Mode.PUNCT, Mode.DIGIT):
                             # if changing from punct or digit to binary mode use U/L as intermediate mode
                             # TODO: update for digit
-                            back_to[y] = 'upper'
-                            cur_seq[y] = cur_seq[x] + ['U/L', '%s/S' % y.upper()[0], 'size']
+                            back_to[y] = Mode.UPPER
+                            cur_seq[y] += [Latch.UPPER, Shift.BINARY, Misc.SIZE]
                         else:
                             back_to[y] = x
-                            cur_seq[y] = cur_seq[x] + ['%s/S' % y.upper()[0], 'size']
-                    else:
-                        if cur_seq[x]:
-                            # if changing from punct or digit mode - use U/L as intermediate mode
-                            # TODO: update for digit
-                            if x in ['punct', 'digit'] and y != 'upper':
-                                cur_seq[y] = cur_seq[x] + ['resume', 'U/L', '%s/L' % y.upper()[0]]
-                                back_to[y] = y
-                            elif x in ['upper', 'lower'] and y == 'punct':
-                                cur_seq[y] = cur_seq[x] + ['M/L', '%s/L' % y.upper()[0]]
-                                back_to[y] = y
-                            elif x == 'mixed' and y != 'upper':
-                                if y == 'punct':
-                                    cur_seq[y] = cur_seq[x] + ['P/L']
-                                    back_to[y] = 'punct'
-                                else:
-                                    cur_seq[y] = cur_seq[x] + ['U/L', 'D/L']
-                                    back_to[y] = 'digit'
-                                continue
+                            cur_seq[y] += [Shift.BINARY, Misc.SIZE]
+                    elif cur_seq[x]:
+                        # if changing from punct or digit mode - use U/L as intermediate mode
+                        # TODO: update for digit
+                        if x in (Mode.PUNCT, Mode.DIGIT) and y != Mode.UPPER:
+                            cur_seq[y] += [Misc.RESUME, Latch.UPPER, Latch[y.name]]
+                        elif x in (Mode.UPPER, Mode.LOWER) and y == Mode.PUNCT:
+                            cur_seq[y] += [Latch.MIXED, Latch[y.name]]
+                        elif x == Mode.MIXED and y != Mode.UPPER:
+                            if y == Mode.PUNCT:
+                                cur_seq[y] += [Latch.PUNCT]
+                                back_to[y] = Mode.PUNCT
                             else:
-                                if x == 'binary':
-                                    # TODO: review this
-                                    # Reviewed by jravallec
-                                    if y == back_to[x]:
-                                        # when return from binary to previous mode, skip mode change
-                                        cur_seq[y] = cur_seq[x] + ['resume']
-                                    elif y == 'upper':
-                                        if back_to[x] == 'lower':
-                                            cur_seq[y] = cur_seq[x] + ['resume', 'M/L', 'U/L']
-                                        if back_to[x] == 'mixed':
-                                            cur_seq[y] = cur_seq[x] + ['resume', 'U/L']
-                                        back_to[y] = 'upper'
-                                    elif y == 'lower':
-                                        cur_seq[y] = cur_seq[x] + ['resume', 'L/L']
-                                        back_to[y] = 'lower'
-                                    elif y == 'mixed':
-                                        cur_seq[y] = cur_seq[x] + ['resume', 'M/L']
-                                        back_to[y] = 'mixed'
-                                    elif y == 'punct':
-                                        if back_to[x] == 'mixed':
-                                            cur_seq[y] = cur_seq[x] + ['resume', 'P/L']
-                                        else:
-                                            cur_seq[y] = cur_seq[x] + ['resume', 'M/L', 'P/L']
-                                        back_to[y] = 'punct'
-                                    elif y == 'digit':
-                                        if back_to[x] == 'mixed':
-                                            cur_seq[y] = cur_seq[x] + ['resume', 'U/L', 'D/L']
-                                        else:
-                                            cur_seq[y] = cur_seq[x] + ['resume', 'D/L']
-                                        back_to[y] = 'digit'
+                                cur_seq[y] += [Latch.UPPER, Latch.DIGIT]
+                                back_to[y] = Mode.DIGIT
+                            continue
+                        elif x == Mode.BINARY:
+                            # TODO: review this
+                            # Reviewed by jravallec
+                            if y == back_to[x]:
+                                # when return from binary to previous mode, skip mode change
+                                cur_seq[y] += [Misc.RESUME]
+                            elif y == Mode.UPPER:
+                                if back_to[x] == Mode.LOWER:
+                                    cur_seq[y] += [Misc.RESUME, Latch.MIXED, Latch.UPPER]
+                                if back_to[x] == Mode.MIXED:
+                                    cur_seq[y] += [Misc.RESUME, Latch.UPPER]
+                            elif y == Mode.LOWER:
+                                cur_seq[y] += [Misc.RESUME, Latch.LOWER]
+                            elif y == Mode.MIXED:
+                                cur_seq[y] += [Misc.RESUME, Latch.MIXED]
+                            elif y == Mode.PUNCT:
+                                if back_to[x] == Mode.MIXED:
+                                    cur_seq[y] += [Misc.RESUME, Latch.PUNCT]
                                 else:
-                                    cur_seq[y] = cur_seq[x] + ['resume', '%s/L' % y.upper()[0]]
-                                    back_to[y] = y
+                                    cur_seq[y] += [Misc.RESUME, Latch.MIXED, Latch.PUNCT]
+                            elif y == Mode.DIGIT:
+                                if back_to[x] == Mode.MIXED:
+                                    cur_seq[y] += [Misc.RESUME, Latch.UPPER, Latch.DIGIT]
+                                else:
+                                    cur_seq[y] += [Misc.RESUME, Latch.DIGIT]
                         else:
-                            # if changing from punct or digit mode - use U/L as intermediate mode
-                            # TODO: update for digit
-                            if x in ['punct', 'digit']:
-                                cur_seq[y] = cur_seq[x] + ['U/L', '%s/L' % y.upper()[0]]
-                                back_to[y] = y
-                            elif x in ['binary', 'upper', 'lower'] and y == 'punct':
-                                cur_seq[y] = cur_seq[x] + ['M/L', '%s/L' % y.upper()[0]]
-                                back_to[y] = y
-                            else:
-                                cur_seq[y] = cur_seq[x] + ['%s/L' % y.upper()[0]]
-                                back_to[y] = y
-        next_len = {
-            'upper': E, 'lower': E, 'mixed': E, 'punct': E, 'digit': E, 'binary': E
-        }
-        next_seq = {
-            'upper': [], 'lower': [], 'mixed': [], 'punct': [], 'digit': [], 'binary': []
-        }
-        possible_modes = []
-        if c in upper_chars:
-            possible_modes.append('upper')
-        if c in lower_chars:
-            possible_modes.append('lower')
-        if c in mixed_chars:
-            possible_modes.append('mixed')
-        if c in punct_chars:
-            possible_modes.append('punct')
-        if c in digit_chars:
-            possible_modes.append('digit')
-        possible_modes.append('binary')
+                            cur_seq[y] += [Misc.RESUME, Latch[y.name]]
+                    else:
+                        # if changing from punct or digit mode - use U/L as intermediate mode
+                        # TODO: update for digit
+                        if x in (Mode.PUNCT, Mode.DIGIT):
+                            cur_seq[y] = [Latch.UPPER, Latch[y.name]]
+                        elif x in (Mode.BINARY, Mode.UPPER, Mode.LOWER) and y == Mode.PUNCT:
+                            cur_seq[y] = [Latch.MIXED, Latch[y.name]]
+                        else:
+                            cur_seq[y] = [Latch[y.name]]
+        next_len = {m:E for m in Mode}
+        next_seq = {m:[] for m in Mode}
+        possible_modes = [m for m in Mode if m == Mode.BINARY or c in code_chars[m]]
         for x in possible_modes:
             # TODO: review this!
-            if back_to[x] == 'digit' and x == 'lower':
-                cur_seq[x] = cur_seq[x] +  ['U/L', 'L/L']
-                cur_len[x] = cur_len[x] + latch_len[back_to[x]][x]
-                back_to[x] = 'lower'
+            if back_to[x] == Mode.DIGIT and x == Mode.LOWER:
+                cur_seq[x] += [Latch.UPPER, Latch.LOWER]
+                cur_len[x] += latch_len[back_to[x]][x]
+                back_to[x] = Mode.LOWER
             # add char to current sequence
             if cur_len[x] + char_size[x] < next_len[x]:
                 next_len[x] = cur_len[x] + char_size[x]
                 next_seq[x] = cur_seq[x] + [c]
-            for y in modes[:-1]:
-                if y == x:
+            for y in Mode:
+                if y == x or y == Mode.BINARY:
                     continue
                 if cur_len[y] + shift_len[y][x] + char_size[x] < next_len[y]:
                     next_len[y] = cur_len[y] + shift_len[y][x] + char_size[x]
-                    next_seq[y] = cur_seq[y] + ['%s/S' % x.upper()[0]] + [c]
+                    next_seq[y] = cur_seq[y] + [Shift[x.name], c]
         # TODO: review this!!!
-        if prev_c and prev_c + c in punct_2_chars:
-            for x in modes:
-                last_mode = ''
-                for char in cur_seq[x][::-1]:
-                    if char.replace('/S', '').replace('/L', '') in abbr_modes:
-                        last_mode = abbr_modes.get(char.replace('/S', '').replace('/L', ''))
-                        break
-                if last_mode == 'punct':
-                    if cur_seq[x][-1] + c in punct_2_chars:
+        if prev_c and bytes((prev_c, c)) in punct_2_chars:
+            for x in Mode:
+                # Will never StopIteration because we must have one S/L already since prev_c is PUNCT
+                last_mode = next(s.value for s in reversed(cur_seq[x]) if s in Latch or s in Shift)
+                if last_mode == Mode.PUNCT:
+                    last_c = cur_seq[x][-1]
+                    if isinstance(last_c, int) and bytes((last_c, c)) in punct_2_chars:
                         if cur_len[x] < next_len[x]:
                             next_len[x] = cur_len[x]
-                            next_seq[x] = cur_seq[x][:-1] + [cur_seq[x][-1] + c]
-        if len(next_seq['binary']) - 2 == 32:
-            next_len['binary'] += 11
-        for i in modes:
-            cur_len[i] = next_len[i]
-            cur_seq[i] = next_seq[i]
+                            next_seq[x] = cur_seq[x][:-1] + [ bytes((last_c, c)) ]
+        if len(next_seq[Mode.BINARY]) - 2 == 32:
+            next_len[Mode.BINARY] += 11
+        cur_len = next_len.copy()
+        cur_seq = next_seq.copy()
         prev_c = c
     # sort in ascending order and get shortest sequence
     result_seq = []
@@ -372,16 +321,16 @@ def find_optimal_sequence(data):
     sizes = {}
     result_seq_len = len(result_seq)
     reset_pos = result_seq_len - 1
-    for i, c in enumerate(result_seq[::-1]):
-        if c == 'size':
+    for i, c in enumerate(reversed(result_seq)):
+        if c == Misc.SIZE:
             sizes[i] = reset_pos - (result_seq_len - i - 1)
             reset_pos = result_seq_len - i
-        elif c == 'resume':
+        elif c == Misc.RESUME:
             reset_pos = result_seq_len - i - 2
     for size_pos in sizes:
         result_seq[len(result_seq) - size_pos - 1] = sizes[size_pos]
     # remove 'resume' tokens
-    result_seq = [x for x in result_seq if x != 'resume']
+    result_seq = [x for x in result_seq if x != Misc.RESUME]
     # update binary sequences' extra sizes
     updated_result_seq = []
     is_binary_length = False
@@ -396,7 +345,7 @@ def find_optimal_sequence(data):
         else:
             updated_result_seq.append(c)
 
-        if c == 'B/S':
+        if c == Shift.BINARY:
             is_binary_length = True
 
     return updated_result_seq
@@ -409,8 +358,8 @@ def optimal_sequence_to_bits(optimal_sequence):
     :return: string with bits
     """
     out_bits = ''
-    mode = 'upper'
-    prev_mode = 'upper'
+    mode = Mode.UPPER
+    prev_mode = Mode.UPPER
     shift = False
     binary = False
     binary_seq_len = 0
@@ -422,7 +371,7 @@ def optimal_sequence_to_bits(optimal_sequence):
         # read one item from sequence
         ch = sequence.pop(0)
         if binary:
-            out_bits += bin(ord(ch))[2:].zfill(char_size.get(mode))
+            out_bits += bin(ch)[2:].zfill(char_size.get(mode))
             binary_index += 1
             # resume previous mode at the end of the binary sequence
             if binary_index >= binary_seq_len:
@@ -436,13 +385,34 @@ def optimal_sequence_to_bits(optimal_sequence):
             mode = prev_mode
             shift = False
         # get mode from sequence character
-        if ch.endswith('/L'):
-            mode = abbr_modes.get(ch.replace('/L', ''))
-        elif ch.endswith('/S'):
-            mode = abbr_modes.get(ch.replace('/S', ''))
+        if ch in Latch or ch in Shift:
+            mode = ch.value
+        if ch in Shift:
             shift = True
+        # handle FLG(n)
+        if ch == Misc.FLG:
+            if not sequence:
+                raise Exception('Expected FLG(n) value')
+            flg_n = sequence.pop(0)
+            if not isinstance(flg_n, numbers.Number) or not 0 <= flg_n <= 7:
+                raise Exception('FLG(n) value must be a number from 0 to 7')
+            if flg_n == 7:
+                raise Exception('FLG(7) is reserved and currently illegal')
+
+            out_bits += bin(flg_n)[2:].zfill(3)
+            if flg_n > 1:
+                # ECI
+                if not sequence:
+                    raise Exception('Expected FLG({}) to be followed by ECI code'.format(flg_n))
+                eci_code = sequence.pop(0)
+                if not isinstance(eci_code, numbers.Number) or not 0 <= eci_code < (10**flg_n):
+                    raise Exception('Expected FLG({}) ECI code to be a number from 0 to {}'.format(flg_n, (10**flg_n) - 1))
+                out_digits = str(eci_code).zfill(flg_n).encode()
+                for ch in out_digits:
+                    index = code_chars[Mode.DIGIT].index(ch)
+                    out_bits += bin(index)[2:].zfill(char_size[Mode.DIGIT])
         # handle binary mode
-        if mode == 'binary':
+        if mode == Mode.BINARY:
             if not sequence:
                 raise Exception('Expected binary sequence length')
             # followed by a 5 bit length
@@ -514,14 +484,14 @@ def find_suitable_matrix_size(data):
     """ Find suitable matrix size
     Raise an exception if suitable size is not found
 
-    :param data: data to encode
+    :param data: string or bytes to encode
     :return: (size, compact) tuple
     """
     optimal_sequence = find_optimal_sequence(data)
     out_bits = optimal_sequence_to_bits(optimal_sequence)
     for (size, compact) in sorted(table.keys()):
         config = get_config_from_table(size, compact)
-        bits = config.get('bits')
+        bits = config.bits
         # error correction percent
         ec_percent = 23  # recommended: 23% of symbol capacity plus 3 codewords
         # calculate minimum required number of bits
@@ -541,7 +511,7 @@ class AztecCode(object):
         If size and compact parameters are None (by default), an
         optimal size and compactness calculated based on the data.
 
-        :param data: data to encode
+        :param data: string or bytes to encode
         :param size: size of matrix
         :param compact: compactness flag
         """
@@ -664,7 +634,7 @@ class AztecCode(object):
         :param data_cw_count: number of data codewords.
         """
         config = get_config_from_table(self.size, self.compact)
-        layers_count = config.get('layers')
+        layers_count = config.layers
         mode_data_values = self.__get_mode_message(layers_count, data_cw_count)
         mode_data_bits = ''.join('{0:04b}'.format(v) for v in mode_data_values)
 
@@ -714,10 +684,10 @@ class AztecCode(object):
         optimal_sequence = find_optimal_sequence(data)
         out_bits = optimal_sequence_to_bits(optimal_sequence)
         config = get_config_from_table(self.size, self.compact)
-        layers_count = config.get('layers')
-        cw_count = config.get('codewords')
-        cw_bits = config.get('cw_bits')
-        bits = config.get('bits')
+        layers_count = config.layers
+        cw_count = config.codewords
+        cw_bits = config.cw_bits
+        bits = config.bits
 
         # error correction percent
         ec_percent = 23  # recommended
